##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################


def ripeomo(problem_data, ptype, fixarray, cc_int, task, pc, sharedata):
    # This subroutine creates a pyomo model to solve the RIPE MINLP
    # Inputs:
    # problem_data  - process data from ripemodel()
    # ptype         - (
    # fixarray      - array denoting valid stoichiometry/mechanism pairings
    # cc_int        - maximum cardinality
    # task          - 0 for presolve, 1 for RIPE call
    # pc            - process condition array from ripemodel()
    # sharedata     - shared data dctionary
    # Outputs:
    # riperes       - dioctionary containing parameter estimates

    aterm, target, sigma, bounds = problem_data
    # import re
    import pyomo.environ as pyo
    # from pyomo.opt import SolverFactory
    # import ripe
    import numpy as np

    gasc = sharedata["gasconst"]
    n, s, r, h = np.shape(aterm)
    # Import max bounds for parameter
    elb, eub, klb, kub = [
        bounds["e"]["min"],
        bounds["e"]["max"],
        bounds["k"]["min"],
        bounds["k"]["max"],
    ]
    if not isinstance(elb, type([])):
        elb = [elb] * r
    if not isinstance(eub, type([])):
        eub = [eub] * r
    if not isinstance(klb, type([])):
        klb = [klb] * r
    if not isinstance(kub, type([])):
        kub = [kub] * r

    # Initialize models
    opt = pyo.SolverFactory("baron")
    model = pyo.ConcreteModel()

    # Set i over observations
    model.i = pyo.RangeSet(n)
    # Set r over reaction mechanisms
    model.r = pyo.RangeSet(r)
    # Set s over species
    model.s = pyo.RangeSet(s)
    # Set h over stoichiometries
    model.h = pyo.RangeSet(h)

    # Define pre-exponential factor and binary variables
    model.k = pyo.Var(model.r, model.h, domain=pyo.NonNegativeReals)
    model.y = pyo.Var(model.r, model.h, domain=pyo.Binary)

    # Add set parameter and subset for elimination of dependent reactions
    # famcons involve reaction stoichiometries of the same order
    if sharedata["addfamcons"]:
        s_small, s_big, ncon, s_fam, nfam = sharedata["sdep"]
        model.fr = pyo.RangeSet(len(nfam))
        model.fcon = pyo.Param(
            model.fr, initialize=dict(((fr), nfam[fr - 1]) for fr in model.fr)
        )
        model.depfam = pyo.Set(
            model.fr, model.h, initialize=dict(((r), s_fam[r - 1]) for r in model.fr)
        )

        def famcon(model, fr):
            return (
                sum(sum(model.y[r, j] for r in model.r) for j in model.depfam[fr])
                <= model.fcon[fr]
            )

        model.famcons = pyo.Constraint(model.fr, rule=famcon)

    # depcons involve reaction stoiciometries of different orders
    if sharedata["adddepcons"]:
        s_small, s_big, ncon, s_fam, nfam = sharedata["sdep"]
        model.dr = pyo.RangeSet(len(ncon))
        model.scon = pyo.Param(
            model.dr, initialize=dict(((dr), ncon[dr - 1]) for dr in model.dr)
        )
        model.depsmall = pyo.Set(
            model.dr,
            within=model.h,
            initialize=dict(((r), s_small[r - 1]) for r in model.dr),
        )
        model.deplarge = pyo.Set(
            model.dr,
            within=model.h,
            initialize=dict(((r), s_big[r - 1]) for r in model.dr),
        )

        def depcon(model, dr):
            return (
                sum(sum(model.y[r, j] for r in model.r) for j in model.depsmall[dr])
                + model.scon[dr]
                * sum(sum(model.y[r, j] for r in model.r) for j in model.deplarge[dr])
                <= model.scon[dr]
            )

        model.depcons = pyo.Constraint(model.dr, rule=depcon)

    # ptype == 'arr' if temperature is included in kwargs and arrhenious relationships will be fit
    if ptype == "arr":
        model.E = pyo.Var(model.r, model.h, domain=pyo.Reals)
        model.T = pyo.Param(
            model.i, initialize=dict(((i), pc["T"][i - 1][0]) for i in model.i)
        )
        model.eub = pyo.Param(
            model.r, initialize=dict(((r), eub[r - 1]) for r in model.r)
        )
        model.elb = pyo.Param(
            model.r, initialize=dict(((r), elb[r - 1]) for r in model.r)
        )
        if "Tref" in pc.keys():
            model.Tref = pyo.Param(rule=pc["Tref"])

    model.aterm = pyo.Param(
        model.i,
        model.s,
        model.r,
        model.h,
        initialize=dict(
            ((i, s, r, h), aterm[i - 1, s - 1, r - 1, h - 1])
            for i in model.i
            for s in model.s
            for r in model.r
            for h in model.h
        ),
    )
    model.target = pyo.Param(
        model.i,
        model.s,
        initialize=dict(
            ((i, s), target[i - 1, s - 1]) for i in model.i for s in model.s
        ),
    )
    model.sigma = pyo.Param(
        model.i,
        model.s,
        initialize=dict(
            ((i, s), sigma[i - 1][s - 1]) for s in model.s for i in model.i
        ),
    )
    model.kub = pyo.Param(model.r, initialize=dict(((r), kub[r - 1]) for r in model.r))
    model.klb = pyo.Param(model.r, initialize=dict(((r), klb[r - 1]) for r in model.r))

    # Define isothermal objective function (no Arrhenious)
    def obj_expression(model):
        return sum(
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(model.k[r, h] * model.aterm[i, s, r, h] for r in model.r)
                        for h in model.h
                    )
                )
                ** 2
                for i in model.i
            )
            / model.sigma[s]
            for s in model.s
        )

    # Define arrhenious objective function
    def arr_expression(model):
        return sum(
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(
                            model.k[r, h]
                            * pyo.exp(-1.0 * model.E[r, h] / (gasc * model.T[i]))
                            * model.aterm[i, s, r, h]
                            for r in model.r
                        )
                        for h in model.h
                    )
                )
                ** 2
                for i in model.i
            )
            / model.sigma[s]
            for s in model.s
        )

    def refarr_expression(model):
        return sum(
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(
                            model.k[r, h]
                            * pyo.exp(
                                -1.0
                                * (model.E[r, h] / gasc)
                                * (1 / (model.T[i]) - 1 / model.Tref)
                            )
                            * model.aterm[i, s, r, h]
                            for r in model.r
                        )
                        for h in model.h
                    )
                )
                ** 2
                for i in model.i
            )
            / model.sigma[s]
            for s in model.s
        )

    # Define additional objectives for wls, these could be combined with a simplified approach but are left separate for clarity for now
    def wls_obj_expression(model):
        return sum(
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(model.k[r, h] * model.aterm[i, s, r, h] for r in model.r)
                        for h in model.h
                    )
                )
                ** 2
                / model.sigma[i, s]
                for i in model.i
            )
            for s in model.s
        )

    # Define arrhenious objective function
    def wls_arr_expression(model):
        return sum(
            sum(
                (
                    (
                        model.target[i, s]
                        - sum(
                            sum(
                                model.k[r, h]
                                * pyo.exp(-model.E[r, h] / (gasc * model.T[i]))
                                * (model.aterm[i, s, r, h])
                                for r in model.r
                            )
                            for h in model.h
                        )
                    )
                    ** 2
                    / model.sigma[i, s]
                )
                for i in model.i
            )
            for s in model.s
        )

    def wls_refarr_expression(model):
        #        return sum(sum((model.target[i,s] - sum(sum(model.k[r,h] * pyo.exp( -1.0 * model.e[r,h] * ( 1.0 / (8.314*model.T[i])))*model.aterm[i,s,r,h] for r in model.r) for h in model.h))**(2) / model.sigma[i,s] for i in model.i) for s in model.s)
        return sum(
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(
                            model.k[r, h]
                            * pyo.exp(
                                -1.0
                                * (model.E[r, h] / gasc)
                                * (1 / (model.T[i]) - 1 / model.Tref)
                            )
                            * model.aterm[i, s, r, h]
                            for r in model.r
                        )
                        for h in model.h
                    )
                )
                ** 2
                / model.sigma[i, s]
                for i in model.i
            )
            for s in model.s
        )

    # Check for temperature dependencies
    if ptype == "arr":
        if "Tref" in pc.keys():
            model.OBJ = pyo.Objective(rule=wls_refarr_expression)
        else:
            model.OBJ = pyo.Objective(rule=wls_arr_expression)
    else:
        model.OBJ = pyo.Objective(rule=wls_obj_expression)

    # Define big-M constraints for k and possibly E
    def bigMub(model, r, h):
        return model.k[r, h] <= model.kub[r] * model.y[r, h]

    def bigMlb(model, r, h):
        return model.k[r, h] >= model.klb[r] * model.y[r, h]

    def bigEub(model, r, h):
        return model.E[r, h] <= model.eub[r] * model.y[r, h]

    def bigElb(model, r, h):
        return model.E[r, h] >= model.elb[r] * model.y[r, h]

    model.ubcon = pyo.Constraint(model.r, model.h, rule=bigMub)
    model.lbcon = pyo.Constraint(model.r, model.h, rule=bigMlb)
    if ptype == "arr":
        model.ubecon = pyo.Constraint(model.r, model.h, rule=bigEub)
        model.lbecon = pyo.Constraint(model.r, model.h, rule=bigElb)

    # Constriant defining one mechanisms per reaction stoichiometry
    def onemech(model, h):
        return sum(model.y[r, h] for r in model.r) <= 1.0

    if sharedata["onemechper"]:
        model.mechcon = pyo.Constraint(model.h, rule=onemech)

    # Cardinality constraint
    def ccon(model):
        return sum(sum(model.y[r, h] for h in model.h) for r in model.r) == cc_int

    # relax equality constraint
    def ccon_relax(model):
        return sum(sum(model.y[r, h] for h in model.h) for r in model.r) <= cc_int

    # hard cardinality vs soft for task 1 vs 0
    if task == 0:
        model.cardcon = pyo.Constraint(rule=ccon_relax)
    else:
        model.cardcon = pyo.Constraint(rule=ccon)

    # Fix array is used to fix binary variables of reactions/mechanisms that should not be considered
    nst, nm = np.shape(fixarray)
    for i in range(nst):
        for j in range(nm):
            if fixarray[i, j] == 0:
                model.y[j + 1, i + 1] = 0
                model.y[j + 1, i + 1].fixed = True

    # initialize results dictionary
    riperes = {}

    if task == 0:
        results = opt.solve(
            model,
            tee=sharedata["showpyomo"],
            options={"DeltaTerm": 1},
            keepfiles=sharedata["keepfiles"],
        )
    else:
        results = opt.solve(
            model,
            tee=sharedata["showpyomo"],
            keepfiles=sharedata["keepfiles"],
            options={
                "MaxTime": sharedata["maxmiptime"],
                "DeltaTerm": sharedata["deltaterm"],
            },
        )
    # Find residual values
    if ptype == "arr":
        if "Tref" in pc.keys():
            sr_list = [
                [
                    (
                        model.target[i, s]
                        - sum(
                            sum(
                                model.k[r, h].value
                                * pyo.exp(
                                    -1.0
                                    * (model.E[r, h].value / gasc)
                                    * (1 / (model.Tref) - 1 / (model.T[i]))
                                )
                                * model.aterm[i, s, r, h]
                                for r in model.r
                            )
                            for h in model.h
                        )
                    )
                    ** 2
                    for s in model.s
                ]
                for i in model.i
            ]
            ssr_list = ssr_list = [
                sum(
                    (
                        model.target[i, s]
                        - sum(
                            sum(
                                model.k[r, h].value
                                * pyo.exp(
                                    -1.0
                                    * (model.E[r, h].value / gasc)
                                    * (1 / (model.Tref) - 1 / (model.T[i]))
                                )
                                * model.aterm[i, s, r, h]
                                for r in model.r
                            )
                            for h in model.h
                        )
                    )
                    ** 2
                    for i in model.i
                )
                for s in model.s
            ]
        else:
            ssr_list = [
                sum(
                    (
                        model.target[i, s]
                        - sum(
                            sum(
                                model.k[r, h].value
                                * pyo.exp(
                                    -1.0 * model.E[r, h].value / (gasc * model.T[i])
                                )
                                * model.aterm[i, s, r, h]
                                for r in model.r
                            )
                            for h in model.h
                        )
                    )
                    ** 2
                    for i in model.i
                )
                for s in model.s
            ]
            sr_list = [
                [
                    (
                        model.target[i, s]
                        - sum(
                            sum(
                                model.k[r, h].value
                                * pyo.exp(
                                    -1.0 * model.E[r, h].value / (gasc * model.T[i])
                                )
                                * model.aterm[i, s, r, h]
                                for r in model.r
                            )
                            for h in model.h
                        )
                    )
                    ** 2
                    for s in model.s
                ]
                for i in model.i
            ]
    else:
        sr_list = [
            [
                (
                    model.target[i, s]
                    - sum(
                        sum(
                            model.k[r, h].value * model.aterm[i, s, r, h]
                            for r in model.r
                        )
                        for h in model.h
                    )
                )
                ** 2
                for s in model.s
            ]
            for i in model.i
        ]
        ssr_list = [
            sum(
                (
                    model.target[i, s]
                    - sum(
                        sum(
                            model.k[r, h].value * model.aterm[i, s, r, h]
                            for r in model.r
                        )
                        for h in model.h
                    )
                )
                ** 2
                for i in model.i
            )
            for s in model.s
        ]

    # Use results to estimate sigma values
    riperes["sigma"] = []
    for ssr in ssr_list:
        riperes["sigma"].append(ssr / np.max((1, float(n - cc_int))))
    riperes["wlsmat"] = []
    for sr in sr_list:
        riperes["wlsmat"].append([1 / np.max([0.000001, s]) for s in sr])
    model.solutions.store_to(results)

    # Populate results dictionary for post processing
    for key in ["ind", "k", "OBJ"]:
        riperes[key] = []
    if ptype == "arr":
        riperes["E"] = []

    # smallnum = 10 ** -6
    riperes["OBJ"] = results.Solution.Objective["OBJ"]["Value"]

    for v in results.Solution.Variable.keys():
        # find active binaries
        if "y" in v:
            # double test currently needed for obtained results
            if (
                results.Solution.Variable[v]["Value"] > 0.99
                and results.Solution.Variable[v]["Value"] < 1.1
            ):
                line = v.split(",")
                line = [line[0].split("[")[-1], line[1].split("]")[0]]
                line = [line[0], line[1]]
                riperes["ind"].append(line)
    # results may be in any order so we need to loop again and check the appropriate indicies
    if task == 0:
        riperes["maxk"] = 0.0
        riperes["maxe"] = 0.0

    for v in results.Solution.Variable.keys():
        line = v.split(",")
        line = [line[0].split("[")[-1], line[1].split("]")[0]]
        line = [line[0], line[1]]
        for inds in riperes["ind"]:
            if inds[0] == line[0] and inds[1] == line[1]:
                if "k" in v:
                    riperes["k"].append(
                        [
                            float(results.Solution.Variable[v]["Value"]),
                            [inds[0], inds[1]],
                        ]
                    )
                    if task == 0:
                        riperes["maxk"] = max(
                            riperes["maxk"],
                            float(results.Solution.Variable[v]["Value"]),
                        )
                if "E" in v:
                    riperes["E"].append(
                        [
                            float(results.Solution.Variable[v]["Value"]),
                            [inds[0], inds[1]],
                        ]
                    )
                    if task == 0:
                        riperes["maxe"] = max(
                            riperes["maxe"],
                            float(results.Solution.Variable[v]["Value"]),
                        )
    return riperes



__all__ = ['alamo', 'doalamo', 'getlabels', 'makelabs', 'addCustomFunctions', 
           'addCustomConstraints', 'addBasisGroups', 'addBasisGroup', 
           'addBasisConstraints', 'addBasisConstraint',
           'get_alamo_version', 'writethis', 'almwriter', 
           'almconfidence', 'almpickle', 'postpickle', 'almunpickle',
           'almerror', 'AlamoError', 'AlamoInputError',
           'almpywriter', 'almcvwriter', 'wrapwriter',
           'almplot', 'mapminmax', 'remapminmax', 
           'data', 'debug', 'deletefile', 'movefile', 'catfile', 
           'copyfile', 'has_alamo', 'allcard', 'almlsq',
           'almlsqjac', 'almfeatmat',
           'ackley', 'branin', 'sixcamel', 'col']

from .doalamo import alamo, doalamo, getlabels, makelabs, addCustomFunctions, \
    addCustomConstraints, addBasisGroups, addBasisGroup, addBasisConstraints, \
    addBasisConstraint, get_alamo_version
from .allcard import allcard, almlsq, almlsqjac, almfeatmat
from .writethis import writethis
from .almwriter import almwriter
from .almconfidence import almconfidence
from .almpickle import almpickle, postpickle, almunpickle
from .almerror import almerror, AlamoError, AlamoInputError
from .almpywriter import almpywriter, almcvwriter, wrapwriter
from .mapminmax import mapminmax
from .remapminmax import remapminmax
from .shared import data, debug
from .almplot import almplot
from .multos import deletefile, movefile, catfile, copyfile, has_alamo
from .examples import sixcamel, ackley, branin

# Initializes all values to be used in the .alm file
import collections as col



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Convenience plotting functions for time-dependent variables.
"""

__author__ = "John Eslick"

import pyomo.environ as pyo
import matplotlib.pyplot as plt

def stitch_dynamic(*args):
    """
    Combine time-indexed Pyomo component values from different models into one
    combined time set. This allows you to use multiple models to simulate
    sections of the time domain, and plot them all together.

    Args:
        Positional arguments (): Multiple Pyomo components indexed by time, or
            time sets

    Returns:
        (list) with the time indexed Pyomo compoent values concatonated for
            plotting
    """
    l = []
    for v in args:
        if isinstance(v, pyo.Set):
            l += [t for t in v]
        else:
            l += [pyo.value(v[t]) for t in v]
    return l

def plot_dynamic(time, y, ylabel, xlabel="time (s)", title=None, legend=None):
    """
    Plot time dependent variables with pyplot.

    Args:
        time (ContinuousSet or list-like): Time index set
        y (list-like of list-likes of Var, Expression, Reference, or float):
            List of quantities to plot (multiple quantities can be plotted). Each
            quantity in the list should be indexed only by time. If you want to
            plot something that is not indexed only by time, you can create a
            Pyomo Reference with the correct indexing.
        ylabel (str): Y-axis label, required
        xlabel (str): X-axis label, default = 'time (s)'
        title (str or None): Plot title, default = None
        legend (list-like of str or None): Legend string for each y,
            default = None

    Returns:
        None
    """
    for i, z in enumerate(y):
        if isinstance(z, (list, tuple)):
            continue # don't need to convert this, because already a list
        y[i] = [pyo.value(z[t]) for t in time]
    for q in y:
        plt.plot(time, q)
    if legend is not None:
        plt.legend(legend)
    if title is not None:
        plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.tight_layout()
    plt.show()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################

from pandas import DataFrame
from collections import OrderedDict
from pyomo.environ import value
from pyomo.network import Arc, Port

from idaes.core.util.exceptions import ConfigurationError

__author__ = "John Eslick, Andrew Lee"


def create_stream_table_dataframe(streams,
                                  true_state=False,
                                  time_point=0,
                                  orient='columns'):
    """
    Method to create a stream table in the form of a pandas dataframe. Method
    takes a dict with name keys and stream values. Use an OrderedDict to list
    the streams in a specific order, otherwise the dataframe can be sorted
    later.

    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        true_state : indicated whether the stream table should contain the
            display variables define in the StateBlock (False, default) or the
            state variables (True).
        time_point : point in the time domain at which to generate stream table
            (default = 0)
        orient : orientation of stream table. Accepted values are 'columns'
            (default) where streams are displayed as columns, or 'index' where
            stream are displayed as rows.

    Returns:
        A pandas DataFrame containing the stream table data.
    """
    stream_attributes = OrderedDict()

    for n in streams.keys():
        try:
            if isinstance(streams[n], Arc):
                # Use destination of Arc, as inlets are more likely (?) to be
                # fully-defined StateBlocks
                sb = _get_state_from_port(streams[n].destination, time_point)
            elif isinstance(streams[n], Port):
                sb = _get_state_from_port(streams[n], time_point)
            else:
                sb = streams[n][time_point]

            if true_state:
                disp_dict = sb.define_state_vars()
            else:
                disp_dict = sb.define_display_vars()

            stream_attributes[n] = {}

            for k in disp_dict:
                for i in disp_dict[k]:
                    if i is None:
                        stream_attributes[n][k] = value(disp_dict[k][i])
                    else:
                        stream_attributes[n][k+" "+str(i)] = \
                            value(disp_dict[k][i])
        except (AttributeError, KeyError):
            raise TypeError(
                    f"Unrecognised component provided in stream argument "
                    f"{streams[n]}. get_stream_table_attributes only "
                    f"supports Arcs, Ports or StateBlocks.")

    return DataFrame.from_dict(stream_attributes, orient=orient)


def stream_table_dataframe_to_string(stream_table, **kwargs):
    """
    Method to print a stream table from a dataframe. Method takes any argument
    understood by DataFrame.to_string
    """
    # Set some default values for keyword arguments
    na_rep = kwargs.pop("na_rep", "-")
    justify = kwargs.pop("justify", "center")
    float_format = kwargs.pop(
            "float_format",
            lambda x: "{:#.5g}".format(x))

    # Print stream table
    return stream_table.to_string(na_rep=na_rep,
                                  justify=justify,
                                  float_format=float_format,
                                  **kwargs)


def _get_state_from_port(port, time_point):
    # Check port for _state_block attribute
    try:
        if len(port._state_block) == 1:
            return port._state_block[0][time_point]
        else:
            return port._state_block[0][time_point, port._state_block[1]]
    except AttributeError:
        # Port was not created by IDAES add_port methods. Return exception for
        # the user to fix.
        raise ConfigurationError(
                f"Port {port.name} does not have a _state_block attribute, "
                f"thus cannot determine StateBlock to use for collecting data."
                f" Please provide the associated StateBlock instead, or use "
                f"the IDAES add_port methods to create the Port.")


def generate_table(blocks, attributes, heading=None):
    """
    Create a Pandas DataFrame that contains a list of user-defined attributes
    from a set of Blocks.

    Args:
        blocks (dict): A dictionary with name keys and BlockData objects for
            values. Any name can be associated with a block. Use an OrderedDict
            to show the blocks in a specific order, otherwise the dataframe can
            be sorted later.
        attributes (list or tuple of strings): Attributes to report from a
            Block, can be a Var, Param, or Expression. If an attribute doesn't
            exist or doesn't have a valid value, it will be treated as missing
            data.
        heading (list or tuple of srings): A list of strings that will be used
            as column headings. If None the attribute names will be used.
    Returns:
        (DataFrame): A Pandas dataframe containing a data table
    """
    if heading is None:
        heading = attributes
    st = DataFrame(columns=heading)
    row = [None]*len(attributes)  # not a big deal but save time on realloc
    for key, s in blocks.items():
        for i, a in enumerate(attributes):
            try:
                v = getattr(s, a, None)
                v = value(v, exception=False)
            except ZeroDivisionError:
                v = None
            row[i] = v
        st.loc[key] = row
    return st



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for saponification property package example.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import (ConcreteModel,
                           Constraint,
                           Param,
                           value,
                           Var)
from idaes.core import (MaterialBalanceType,
                        EnergyBalanceType,
                        MaterialFlowBasis)

from idaes.property_models.examples.saponification_thermo import (
    SaponificationParameterBlock, SaponificationStateBlock)

from idaes.core.util.testing import get_default_solver


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_default_solver()


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        return model

    def test_config(self, model):
        assert len(model.params.config) == 1

    def test_build(self, model):
        assert model.params.state_block_class is SaponificationStateBlock

        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i == "Liq"

        assert len(model.params.component_list) == 5
        for i in model.params.component_list:
            assert i in ['H2O',
                         'NaOH',
                         'EthylAcetate',
                         'SodiumAcetate',
                         'Ethanol']

        assert isinstance(model.params.cp_mol, Param)
        assert value(model.params.cp_mol) == 75.327

        assert isinstance(model.params.dens_mol, Param)
        assert value(model.params.dens_mol) == 55388

        assert isinstance(model.params.pressure_ref, Param)
        assert value(model.params.pressure_ref) == 101325

        assert isinstance(model.params.temperature_ref, Param)
        assert value(model.params.temperature_ref) == 298.15


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.state_block_class(
                [1],
                default={"parameters": model.params})

        return model

    def test_build(self, model):
        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].conc_mol_comp, Var)
        assert len(model.props[1].conc_mol_comp) == 5
        for i in model.props[1].conc_mol_comp:
            assert value(model.props[1].conc_mol_comp[i]) == 100

        assert isinstance(model.props[1].conc_water_eqn, Constraint)
        assert len(model.props[1].flow_vol) == 1

    def test_build_defined_state(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.state_block_class(
                [1],
                default={"parameters": model.params,
                         "defined_state": True})

        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].conc_mol_comp, Var)
        assert len(model.props[1].conc_mol_comp) == 5
        for i in model.props[1].conc_mol_comp:
            assert value(model.props[1].conc_mol_comp[i]) == 100

        assert not hasattr(model.props[1], "conc_water_eqn")

    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_flow_terms(p, j) == (
                        model.props[1].flow_vol *
                        model.props[1].conc_mol_comp[j])

    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_enthalpy_flow_terms(p) == (
                    model.props[1].flow_vol*model.props[1]._params.dens_mol *
                    model.props[1]._params.cp_mol*(
                            model.props[1].temperature -
                            model.props[1]._params.temperature_ref))

    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert model.props[1].get_material_density_terms(p, j) == (
                        model.props[1].conc_mol_comp[j])

    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert model.props[1].get_energy_density_terms(p) == (
                    model.props[1]._params.dens_mol *
                    model.props[1]._params.cp_mol*(
                            model.props[1].temperature -
                            model.props[1]._params.temperature_ref))

    def test_default_material_balance_type(self, model):
        assert model.props[1].default_material_balance_type() == \
            MaterialBalanceType.componentPhase

    def test_default_energy_balance_type(self, model):
        assert model.props[1].default_energy_balance_type() == \
            EnergyBalanceType.enthalpyTotal

    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == \
            MaterialFlowBasis.molar

    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_vol",
                         "conc_mol_comp",
                         "temperature",
                         "pressure"]

    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_vol",
                         "conc_mol_comp",
                         "temperature",
                         "pressure"]

    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["Volumetric Flowrate",
                         "Molar Concentration",
                         "Temperature",
                         "Pressure"]

    def test_model_check_none(self, model, caplog):
        assert model.props[1].model_check() is None
        assert 'Temperature set below lower bound' not in caplog.text
        assert 'Temperature set above upper bound' not in caplog.text
        assert 'Pressure set below lower bound' not in caplog.text
        assert 'Pressure set above upper bound' not in caplog.text

    def test_model_check_low_T(self, model, caplog):
        model.props[1].temperature.value = 200
        assert model.props[1].model_check() is None
        assert 'Temperature set below lower bound' in caplog.text
        assert 'Temperature set above upper bound' not in caplog.text
        assert 'Pressure set below lower bound' not in caplog.text
        assert 'Pressure set above upper bound' not in caplog.text

    def test_model_check_high_T(self, model, caplog):
        model.props[1].temperature.value = 350
        assert model.props[1].model_check() is None
        assert 'Temperature set below lower bound' not in caplog.text
        assert 'Temperature set above upper bound' in caplog.text
        assert 'Pressure set below lower bound' not in caplog.text
        assert 'Pressure set above upper bound' not in caplog.text
        # Reset temeprature
        model.props[1].temperature.value = 298.15

    def test_model_check_low_P(self, model, caplog):
        model.props[1].pressure.value = 1e2
        assert model.props[1].model_check() is None
        assert 'Temperature set below lower bound' not in caplog.text
        assert 'Temperature set above upper bound' not in caplog.text
        assert 'Pressure set below lower bound' in caplog.text
        assert 'Pressure set above upper bound' not in caplog.text

    def test_model_check_high_P(self, model, caplog):
        model.props[1].pressure.value = 1e7
        assert model.props[1].model_check() is None
        assert 'Temperature set below lower bound' not in caplog.text
        assert 'Temperature set above upper bound' not in caplog.text
        assert 'Pressure set below lower bound' not in caplog.text
        assert 'Pressure set above upper bound' in caplog.text
        # Reset pressure
        model.props[1].pressure.value = 101325

    def test_initialize(self, model):
        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

        model.props.initialize(hold_state=False, outlvl=1)

        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

    def test_initialize_hold(self, model):
        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

        flags = model.props.initialize(hold_state=True)

        assert model.props[1].flow_vol.fixed
        assert model.props[1].temperature.fixed
        assert model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert model.props[1].conc_mol_comp[i].fixed

        model.props.release_state(flags, outlvl=1)

        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
# =============================================================================

from pyomo.environ import (Constraint,
                           Expression,
                           Var,
                           Param,
                           exp,
                           log10)

def global_costing_parameters(self, year=None):
    if year is None:
        year='2018'
    # Cost index $/year (method argument or 2018 default)
    ce_index_dic = {'2019':680, '2018':671.1, '2017':567.5, '2016':541.7,
                '2015':556.8, '2014':576.1, '2013':567.3, '2012':584.6,
                '2011':585.7, '2010':550.8}
    
    self.CE_index = Param(mutable = True, initialize = ce_index_dic[year],
                          doc= 'CE cost index $ year')

def hx_costing(self, hx_type='U-tube', FM='stain_steel', 
                 FL = '12ft'):
    '''
    Heat exchanger costing method
    Source: Process and Product Design Principles: Synthesis, Analysis, 
    and Evaluation
    Seider, Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons
    Chapter 22. Cost Accounting and Capital Cost Estimation
    22.2 Cost Indexes and Capital Investment
    
    This method computes the purchase cost (CP) for a shell and tube heat
    exchanger (Eq. 22.43), the model computes the base cost (CB for 4 type 
    of heat exchangers, such as floating head, fixed head, U-tube, and 
    Kettle vaporizer), construction material factor (FM), pressure design 
    factor (FP), and tube length correction factor (FL), using CE base cost 
    index of 500. 
    
    Cp = FP*FM*FL*CB
    
    Args:
        hx_type : 'floating_head', 'fixed_head', 'U-tube'*, 'Kettle_vap'
        
        material factor (FM): 'stain_steel'*, 'carb_steel'
        
        tube length (FL): '8ft'*, '12ft', '16ft', '20ft'
        
        * --> default options
    '''
    # ------------------------ Heat Exchanger cost ------------------------
    # heat exchanger cost
    self.base_cost = Var(initialize = 1e5,
                         bounds=(0, 1e12),
                         doc='Heat Exchanger Base Cost CB cost in $')
    self.purchase_cost = Var(initialize =1e4,
                             bounds = (0,1e12),
                             doc = 'Heat Exchanger Purchase Cost CP in $')
    self.FP = Var(initialize = 100,
                  doc = 'hx Pressure factor')
    self.FL = Param(mutable = True, initialize = 1.12,
                    doc='hx tube length correction factor')
    self.FM = Param(mutable = True, initialize = 3,
                    doc= 'construction material correction factor')
    
    self.hx_os = Param(mutable = True, initialize = 1.1,
                       doc= 'hx oversize factor 1.1 to 1.5')

    # select length correction factor 
    c_fl = {'8ft':1.25, '12ft':1.12, '16ft':1.05, '20ft':1.00}
    self.FL = c_fl[FL]

    # select construction material (stainless steel default)
    c_material = {'stain_steel':3, 'carbon_steel':2}
    self.FM = c_material[FM]
    # ToDo (9/17/2019):FM can be calculated for advanced materials Eq.22.44 
    # i.e. Carbon Steel/Cr-Mo steel, Cr-Mo steel/Cr-Mo steel, Monel/Monel, 
    # Titanium/titanium, etc.
    


    #--------------------------------------------------
    # base cost calculation
#       # select heat exchanger type:
    alf1 = {'floating_head':11.9052,'fixed_head':11.2927,
            'U-tube':11.3852,'Kettle_vap':12.2052}
    alf2 = {'floating_head':0.8709,'fixed_head':0.8228,
            'U-tube':0.9186,'Kettle_vap':0.8709}
    alf3 = {'floating_head':0.09005,'fixed_head':0.09861,
            'U-tube':0.09790,'Kettle_vap':0.09005}
    
    if (self.parent_block().config.tube.property_package.get_metadata().
            default_units['length']) == 'm':
        area = self.parent_block().area*10.7639
    elif (self.parent_block().config.tube.property_package.get_metadata().
            default_units['length']) == 'ft':
        area = self.area
    else:
        raise Exception('area units not suported')

    def hx_cost_rule(self):
        return self.base_cost == self.FL * exp(alf1[hx_type]
                    - alf2[hx_type]*log10(area*self.hx_os)
                    + alf3[hx_type]*log10(area*self.hx_os)**2)
    self.base_cost_eq = Constraint(rule=hx_cost_rule)
    
    #------------------------------------------------------
    # Pressure factor calculation
    # doublecheck units (higher pressure fluid should be tube side)
    if (self.parent_block().config.tube.property_package.get_metadata().
            properties['pressure']['units']) == 'Pa':
        pressure = self.parent_block().tube.properties_in[0].pressure*14.69/1.01325e5
    elif (self.parent_block().config.tube.property_package.get_metadata().
            properties['pressure']['units']) == 'psig':
        pressure = self.tube.properties_in[0].pressure

    #units must be in psig 
    #(ToDo(9/17/2019): check inlet pressure to select correlation)
    def hx_P_factor(self):
        # eq valid from 600 pisg to 3000 psig
        #    return self.hx_FP == 0.8510 + 0.1292*(pressure/600) 
        #                                + 0.0198*(pressure/600)**2
        # eq valid from 100 pisg to 2000 psig
        return self.FP == 0.9803 + (0.018*(pressure*14.69/1.01325e5/100)
                                    + 0.0017*(pressure*14.69/1.01325e5/100)**2)
    self.p_factor_eq = Constraint(rule=hx_P_factor)


    #---------------------------------------------------------
    # purchase cost equation
    def hx_CP_rule(self):
        return self.purchase_cost == (self.FP*self.FM*self.FL*
                                      (self.parent_block().flowsheet().costing.CE_index/500)*self.base_cost)
    self.cp_cost_eq = Constraint(rule=hx_CP_rule)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Importing thermodynamic data, specific structures for text files
"""

import numpy as np


filename, sampleRatio = "", ""
Values, DVValues, DLValues, CPValues, PVTValues, PVTSamples, CP0Values, BValues, CVValues, CVValuesn, SNDValues = (
    [],
    [],
    [],
    [],
    [],
    [],
    [],
    [],
    [],
    [],
    [],
)
InSatValues = []
PVTindexes, CVindexes, CPindexes, SNDindexes = [], [], [], []
isothermIndex = 0

critT, critD, critP, acc, R, M, Rm = [0, 0, 0, 0, 0, 0, 0]


def molData(fluidData, RVal):
    """
    Molecular data passed to the module
    """
    global critT, critP, critD, M, triple, acc, R, Rm
    (critT, critP, critD, M, triple, acc) = fluidData
    R = RVal
    Rm = RVal


def regionsOfData(molecule, DataValues, PVT=False, CV=False):
    """
    Organization of data into regions
    """

    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = [], [], [], [], [], []

    if PVT:
        for P, D, T in DataValues:
            T = float(T)
            P = float(P)
            D = float(D)

            vals = [P, D, T]
            if T < float(critT):
                if D < float(critD):
                    Reg1.append(vals)
                else:
                    Reg2.append(vals)
            else:
                if (
                    T / float(critT) < 1.1
                    and T / float(critT) > 0.98
                    and D / float(critD) > 0.7
                    and D / float(critD) < 1.4
                ):
                    Reg3.append(vals)
                else:
                    Rho = D / float(critD)
                    if Rho < 0.6:
                        Reg4.append(vals)
                    elif Rho < 1.5:
                        Reg5.append(vals)
                    else:
                        Reg6.append(vals)
    elif CV:
        for D, T, CV in CVValues:
            T = float(T)
            CV = float(CV)
            D = float(D)

            vals = [D, T, CV]
            if D > 0:
                if T < float(critT):
                    if D < float(critD):
                        Reg1.append(vals)
                    else:
                        Reg2.append(vals)
                else:
                    if (
                        T / float(critT) < 1.1
                        and T / float(critT) > 0.98
                        and D / float(critD) > 0.7
                        and D / float(critD) < 1.4
                    ):
                        Reg3.append(vals)
                    else:
                        Rho = D / float(critD)
                        if Rho < 0.6:
                            Reg4.append(vals)
                        elif Rho < 1.5:
                            Reg5.append(vals)
                        else:
                            Reg6.append(vals)
    else:
        for D, T, Prop in DataValues:
            T = float(T)
            Prop = float(Prop)
            D = float(D)

            vals = [D, T, Prop]
            if T < float(critT):
                if D < float(critD):
                    Reg1.append(vals)
                else:
                    Reg2.append(vals)
            else:
                if (
                    T / float(critT) < 1.1
                    and T / float(critT) > 0.98
                    and D / float(critD) > 0.7
                    and D / float(critD) < 1.4
                ):
                    Reg3.append(vals)
                else:
                    Rho = D / float(critD)
                    if Rho < 0.6:
                        Reg4.append(vals)
                    elif Rho < 1.5:
                        Reg5.append(vals)
                    else:
                        Reg6.append(vals)

    return Reg1, Reg2, Reg3, Reg4, Reg5, Reg6


def sampleData(Regions, ratio):
    """
    Sampling of the data regions
    """
    global isothermIndex
    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = Regions
    Indexes = []

    ind1, ind2, ind3, ind4, ind5, ind6 = 0, 0, 0, 0, 0, 0

    ind1_size = int(len(Reg1) / ratio)
    if len(Reg1) < ratio:
        ind1_size = len(Reg1)
    ind2_size = int(len(Reg2) / ratio)
    if len(Reg2) < ratio:
        ind2_size = len(Reg2)
    ind3_size = int(len(Reg3) / ratio)
    if len(Reg3) < ratio:
        ind3_size = len(Reg3)
    ind4_size = int(len(Reg4) / ratio)
    if len(Reg4) < ratio:
        ind4_size = len(Reg4)
    ind5_size = int(len(Reg5) / ratio)
    if len(Reg5) < ratio:
        ind5_size = len(Reg5)
    ind6_size = int(len(Reg6) / ratio)
    if len(Reg6) < ratio:
        ind6_size = len(Reg6)

    ind1 = np.random.choice(len(Reg1), size=ind1_size, replace=False)
    ind1 = np.asarray(ind1)
    Reg1s = []
    for i in ind1:
        Reg1s.append(Reg1[i])

    ind2 = np.random.choice(len(Reg2), size=ind2_size, replace=False)
    ind2 = np.asarray(ind2)
    Reg2s = []
    for i in ind2:
        Reg2s.append(Reg2[i])

    ind3 = np.random.choice(len(Reg3), size=ind3_size, replace=False)
    ind3 = np.asarray(ind3)
    Reg3s = []
    for i in ind3:
        Reg3s.append(Reg3[i])

    ind4 = np.random.choice(len(Reg4), size=ind4_size, replace=False)
    ind4 = np.asarray(ind4)
    Reg4s = []
    for i in ind4:
        Reg4s.append(Reg4[i])

    ind5 = np.random.choice(len(Reg5), size=ind5_size, replace=False)
    ind5 = np.asarray(ind5)
    Reg5s = []
    for i in ind5:
        Reg5s.append(Reg5[i])

    ind6 = np.random.choice(len(Reg6), size=ind6_size, replace=False)
    ind6 = np.asarray(ind6)
    Reg6s = []
    for i in ind6:
        Reg6s.append(Reg6[i])

    # ind = np.concatenate((ind1, ind2, ind3, ind4, ind5, ind6))

    Total = 0
    for x in [
        len(Reg1s),
        len(Reg2s),
        len(Reg3s),
        len(Reg4s),
        len(Reg5s),
        len(Reg6s),
    ]:
        Total = Total + x
        Indexes.append(Total)

    if len(Reg1s) > 0:
        Reg1s.sort(key=lambda row: row[2:], reverse=False)
        # PUT BACK  AFTER CHECK ENGLE
        try:
            isothermIndex = next(
                i for i, x in enumerate(Reg1s) if x[2] > float(critT) - 20
            )
        except Exception:
            pass

    Data= []
    [Data.append(x) for x in Reg1s[:]]
    [Data.append(x) for x in Reg2s[:]]
    [Data.append(x) for x in Reg3s[:]]
    [Data.append(x) for x in Reg4s[:]]
    [Data.append(x) for x in Reg5s[:]]
    [Data.append(x) for x in Reg6s[:]]

    return Data, Indexes


def PVT(molecule, sample=False, ratio=5):
    """
    Import pressure-volume-temperature data
    """

    global PVTValues, PVTindexes, isothermIndex
    PVTValues, PVTindexes, isothermIndex = [], [], 0

    rawfile = molecule + "PVT.txt"
    dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i > 1:
            line = line.strip()
            b = line.split()
            PVTValues.append(b[0:3])
            i += 1
        else:
            i += 1

    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = regionsOfData(molecule, PVTValues, PVT=True)

    if sample:
        PVTValues, PVTindexes = sampleData((Reg1, Reg2, Reg3, Reg4, Reg5, Reg6), ratio)
    else:
        Reg1.sort(key=lambda row: row[2:], reverse=False)

        if len(Reg1) > 0:
            try:
                isothermIndex = next(
                    i for i, x in enumerate(Reg1) if x[2] > (float(critT) - 30)
                )
            except Exception:
                pass

        Total = 0
        for x in [len(Reg1), len(Reg2), len(Reg3), len(Reg4), len(Reg5), len(Reg6)]:
            Total = Total + x
            PVTindexes.append(Total)

        PVTValues = []
        [PVTValues.append(x) for x in Reg1[:]]
        [PVTValues.append(x) for x in Reg2[:]]
        [PVTValues.append(x) for x in Reg3[:]]
        [PVTValues.append(x) for x in Reg4[:]]
        [PVTValues.append(x) for x in Reg5[:]]
        [PVTValues.append(x) for x in Reg6[:]]


def CP(molecule, sample=False, ratio=5):
    """
    Import isobaric heat capacity data
    """
    global CPValues, CPindexes
    CPValues, CPindexes = [], []

    rawfile = molecule + "CP.txt"
    dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            CPValues.append(b[0:3])
        else:
            i += 1

    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = regionsOfData(molecule, CPValues)

    if sample:
        CPValues, CPindexes = sampleData((Reg1, Reg2, Reg3, Reg4, Reg5, Reg6), ratio)

    else:
        Total = 0
        for x in [len(Reg1), len(Reg2), len(Reg3), len(Reg4), len(Reg5), len(Reg6)]:
            if x == 0:
                CPindexes.append(0)
            else:
                Total = Total + x
                CPindexes.append(Total)

        CPValues = []
        [CPValues.append(x) for x in Reg1[:]]
        [CPValues.append(x) for x in Reg2[:]]
        [CPValues.append(x) for x in Reg3[:]]
        [CPValues.append(x) for x in Reg4[:]]
        [CPValues.append(x) for x in Reg5[:]]
        [CPValues.append(x) for x in Reg6[:]]


def CV(molecule, sample=False, ratio=5):
    """
    Import isochoric heat capacity
    """
    global CVValues, CVindexes
    CVValues, CVindexes = [], []

    rawfile = molecule + "CV.txt"
    dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            CVValues.append(b[0:3])
        else:
            i += 1

    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = regionsOfData(molecule, CVValues, CV=True)
   
    if sample:
        CVValues, CVindexes = sampleData((Reg1, Reg2, Reg3, Reg4, Reg5, Reg6), ratio)
    else:
        Total = 0
        for x in [len(Reg1), len(Reg2), len(Reg3), len(Reg4), len(Reg5), len(Reg6)]:
            if x == 0:
                CVindexes.append(0)
            else:
                Total = Total + x
                CVindexes.append(Total)

        CVValues = []
        [CVValues.append(x) for x in Reg1[:]]
        [CVValues.append(x) for x in Reg2[:]]
        [CVValues.append(x) for x in Reg3[:]]
        [CVValues.append(x) for x in Reg4[:]]
        [CVValues.append(x) for x in Reg5[:]]
        [CVValues.append(x) for x in Reg6[:]]


def SND(molecule, sample=False, ratio=5):
    """
    Import speed of sound data
    """

    global SNDValues, SNDindexes
    SNDValues, SNDindexes = [], []

    rawfile = molecule + "SND.txt"
    dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            SNDValues.append(b[0:3])
            i += 1
        else:
            i += 1

    Reg1, Reg2, Reg3, Reg4, Reg5, Reg6 = regionsOfData(molecule, SNDValues)

    if sample:
        SNDValues, SNDindexes = sampleData((Reg1, Reg2, Reg3, Reg4, Reg5, Reg6), ratio)
    else:
        Total = 0
        for x in [len(Reg1), len(Reg2), len(Reg3), len(Reg4), len(Reg5), len(Reg6)]:
            if x == 0:
                SNDindexes.append(0)
            else:
                Total = Total + x
                SNDindexes.append(Total)

        SNDValues = []
        [SNDValues.append(x) for x in Reg1[:]]
        [SNDValues.append(x) for x in Reg2[:]]
        [SNDValues.append(x) for x in Reg3[:]]
        [SNDValues.append(x) for x in Reg4[:]]
        [SNDValues.append(x) for x in Reg5[:]]
        [SNDValues.append(x) for x in Reg6[:]]


def CP0(molecule):
    """
    Import ideal isobaric heat capacity
    """
    rawfile = molecule + "CP0.RAW"
    dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i > 1:
            line = line.strip()
            b = line.split()
            CP0Values.append(b[0:2])
            i += 1
        else:
            i += 1


def DL(molecule):
    """
    Import saturated liquid density
    """
    rawfile = ""
    dataFile = ""
    try:
        rawfile = molecule + "DL.RAW"
        dataFile = open(filename + "/" + rawfile)
    except Exception:
        rawfile = molecule + "DL.txt"
        dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            DLValues.append(b[0:2])
            i += 1
        else:
            i += 1


def DV(molecule):
    """
    Import of saturated vapor density
    """
    rawfile = ""
    dataFile = ""
    try:
        rawfile = molecule + "DV.RAW"
        dataFile = open(filename + "/" + rawfile)
    except Exception:
        rawfile = molecule + "DV.txt"
        dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            DVValues.append(b[0:2])
            i += 1
        else:
            i += 1


def PV(molecule):
    """
    Import saturated vapor pressure
    """
    rawfile = ""
    dataFile = ""
    try:
        rawfile = molecule + "PV.RAW"
        dataFile = open(filename + "/" + rawfile)
    except Exception:
        rawfile = molecule + "PV.txt"
        dataFile = open(filename + "/" + rawfile)
    i = 0
    for line in dataFile:
        if i != 0 and i != 1:
            line = line.strip()
            b = line.split()
            Values.append(b[0:2])
            i += 1
        else:
            i += 1


# def InSat(molecule):
#     # Cubic-ness
#     Tsat = float(critT) * 0.75
#     DL = SoaveDensity.Sat_Liq_Density(Tsat)
#     DV = SoaveDensity.Sat_Vap_Density(Tsat)
#     for d in np.linspace(DV, DL, 10):
#         Dc = d / float(critD)
#         InSatValues.append([Dc, Tsat])
#     return Tsat, DV, DL
